;;; helm-core-pkg.el --- define helm-core for package.el

(define-package "helm-core" "3.0"
  "Development files for Helm"
  '((emacs "24.4")
    (async "1.9.3"))
  :url "https://emacs-helm.github.io/helm/")

;; Local Variables:
;; no-byte-compile: t
;; End:
